/*
 * This should setup enough memory to relocate the kernel and use fbvideo
 * routines during the system startup phase.
 *
 * By Jon Masters <jcm@jonmasters.org>.
 *
 */

#include <const.h>
#include <hal/ppc/processor.h>
#include <hal/fbvideo/fbvideo_table.h>
#include <mm/mmu.h>

extern unsigned long fbvideo_BATU;
extern unsigned long fbvideo_BATL;

unsigned long system_BATU;
unsigned long system_BATL;

extern fbvideo_info_t fbvideo_default;
extern unsigned long _start;

extern void fbvideo_clear(unsigned long colour);
extern void fbvideo_write(const char *message);
extern void fbvideo_write_hex(unsigned long value);

extern void sysinfo_display(void);

extern unsigned long physical_start_address;

extern void ppc_mmu_flush_tlbs(void);

void setup_BATs(void) {

  /* PowerPC defines a number of Block Address Translation registers.
   * These registers are used to translate large blocks of the virtual
   * address space in to physically contiguous regions of memory. During
   * system startup it is necessary to map both the kernel and the main
   * framebuffer device which will be used for console output.
   */

  /* unsigned long fbvideo_ibatu, fbvideo_ibatl; Unused now. */
  unsigned long fbvideo_dbatu, fbvideo_dbatl;
  unsigned long system_ibatu, system_ibatl;
  unsigned long system_dbatu, system_dbatl;

  /* unsigned long *system_text_ptr; */
  /* unsigned long *system_base_ptr; */
  
  fbvideo_BATU = CONSOLE_BASE | (BAT32_LENGTH_32MB<<2) | BAT32_VS;
  fbvideo_BATL = fbvideo_default.address |
    (BAT32_WIMG_I | BAT32_WIMG_G | BAT32_PP_RW);

  system_BATU  = SYSTEM_BASE | (BAT32_LENGTH_32MB<<2) | BAT32_VS;
  system_BATL  = SYSTEM_ZERO | (BAT32_WIMG_I | BAT32_WIMG_G | BAT32_PP_RW);

  fbvideo_write("setup_BATs: Programming system memory BATs\n");
  fbvideo_write("setup_BATs: Invalidating old BAT entries...");

  /* Assume we have only four I and D BATs in use. In reality there
   * are eight in the MPC7455 however we ignore the newer additions
   * since everything should backwardly compatible.
   */

  __asm__ __volatile__ ("mtspr 0x210,0"); /* Invalidate IBAT0. */
  __asm__ __volatile__ ("mtspr 0x212,0"); /* Invalidate IBAT1. */
  __asm__ __volatile__ ("mtspr 0x214,0"); /* Invalidate IBAT2. */
  __asm__ __volatile__ ("mtspr 0x216,0"); /* Invalidate IBAT3. */
  __asm__ __volatile__ ("mtspr 0x218,0"); /* Invalidate DBAT0. */
  __asm__ __volatile__ ("mtspr 0x21A,0"); /* Invalidate DBAT1. */
  __asm__ __volatile__ ("mtspr 0x21C,0"); /* Invalidate DBAT2. */
  __asm__ __volatile__ ("mtspr 0x21E,0"); /* Invalidate DBAT3. */

  fbvideo_write("done\n");

  /* Mental note to rewrite this to use constants rather than hard coded
   * register addresses please jcm. I just love the PowerPC documentation.
   */

  /* Copy the BATs to the required registers now. */
  
  __asm__ __volatile__ ("mtspr 0x21E,%0"
		        :
			: "r" (fbvideo_BATU) ); /* DBAT3 Upper. */
  __asm__ __volatile__ ("mtspr 0x21F,%0"
		        :
			: "r" (fbvideo_BATL) ); /* DBAT3 Lower. */
  __asm__ __volatile__ ("sync");
  __asm__ __volatile__ ("mtspr 0x210,%0"
			:
			: "r" (system_BATU) );  /* IBAT0 Upper. */
  __asm__ __volatile__ ("mtspr 0x211,%0"
			:
			: "r" (system_BATL) );  /* IBAT0 Lower. */
  __asm__ __volatile__ ("sync");
  __asm__ __volatile__ ("mtspr 0x218,%0"
			:
			: "r" (system_BATU) );  /* DBAT0 Upper. */
  __asm__ __volatile__ ("mtspr 0x219,%0"
			:
			: "r" (system_BATL) );  /* DBAT0 Lower. */

  /* Finished copying so now we synchronise. */

  __asm__ __volatile__ ("sync"); /* Context synchronisation. */
  __asm__ __volatile__ ("isync"); /* Does not hurt to be safe. */

  /* Flush TLBs. */
  ppc_mmu_flush_tlbs();
  
  /* Now be very paranoid and check what we just did. */

  __asm__ __volatile__ ("mfspr %0,0x21E"
			: "=r" (fbvideo_dbatu)); /* Remember this is temp. */
  __asm__ __volatile__ ("mfspr %0,0x21F"
			: "=r" (fbvideo_dbatl)); /* Remember this is temp. */
  __asm__ __volatile__ ("mfspr %0,0x210"
			: "=r" (system_ibatu));  /* Remember this is temp. */
  __asm__ __volatile__ ("mfspr %0,0x211"
			: "=r" (system_ibatl));  /* Remember this is temp. */
  __asm__ __volatile__ ("mfspr %0,0x218"
			: "=r" (system_dbatu));  /* Remember this is temp. */
  __asm__ __volatile__ ("mfspr %0,0x219"
			: "=r" (system_dbatl));  /* Remember this is temp. */

  fbvideo_write("done\n");
  if ((fbvideo_dbatu == fbvideo_BATU) && (fbvideo_dbatl == fbvideo_BATL)) {
    fbvideo_write("setup_pmac: Video Framebuffer BATs match.\r\n");
  } else {
    fbvideo_write("setup_pmac: Video Framebuffer values differ.\r\n");
    fbvideo_write("setup_pmac: [fbvideo] DBATU: ");
    fbvideo_write_hex(fbvideo_dbatu);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac: [fbvideo] DBATL: ");
    fbvideo_write_hex(fbvideo_dbatl);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac: The values programmed were:\r\n");
    fbvideo_write("setup_pmac: [fbvideo] DBATU: ");
    fbvideo_write_hex(fbvideo_BATU);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac: [fbvideo] DBATL: ");
    fbvideo_write_hex(fbvideo_BATL);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac:  [system] IBATU: ");
    fbvideo_write_hex(system_ibatu);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac:  [system] IBATL: ");
    fbvideo_write_hex(system_ibatl);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac:  [system] DBATU: ");
    fbvideo_write_hex(system_dbatu);
    fbvideo_write("\r\n");
    fbvideo_write("setup_pmac:  [system] DBATL: ");
    fbvideo_write_hex(system_dbatl);
    fbvideo_write("\r\n");
    /*    prom_write(prom_stdout,"System halted.\n");
     * moo_cow:
     * goto moo_cow;
     */
    /* fbvideo_default.address = CONSOLE_BASE; */

    fbvideo_write("I am going to modify fbvideo_default.address: ");

    /* Although we have mapped a BAT for the framebuffer it will likely be
     * the case that it is not nicely aligned on the boundary specified.
     * Therefore modify the address to be correct now...
     */

    /* Arrrgh...a giant moo.
     * fbvideo_default.address = CONSOLE_BASE +
     * (fbvideo_default.address & ~0xFF000000UL);
     */

    fbvideo_write_hex(CONSOLE_BASE +
		      (fbvideo_default.address & ~0xFF000000UL));
    fbvideo_write("\n");

    fbvideo_write("The value at physical_start_address: ");
    fbvideo_write_hex(*((unsigned long *)physical_start_address));
    fbvideo_write("\r\n");
    fbvideo_write("The value at SYSTEM_BASE: ");
    fbvideo_write_hex(*((unsigned long *)SYSTEM_BASE));
    fbvideo_write("\r\n");
  }
}
